/*
 * Network Block Device - server
 *
 * Copyright 1996-1998 Pavel Machek, distribute under GPL
 *  <pavel@atrey.karlin.mff.cuni.cz>
 *
 * Version 1.0 - hopefully 64-bit-clean
 * Version 1.1 - merging enhancements from Josh Parsons, <josh@coombs.anu.edu.au>
 * Version 2.1 - enhancements from Peter Breuer, <ptb@it.uc3m.es>
 * Version 2.2 - SSL additions by Andres Marin, <amarin@it.uc3m.es>
 *             - more improvements from ptb, also negotiate cleanups
 *          and improvements from James MacLean <macleajb@EDnet.NS.CA>
 */

#define _LARGEFILE64_SOURCE

#include <sys/socket.h>
#include <netinet/tcp.h>
#include <netinet/in.h>         /* sockaddr_in, htons, in_addr */
#include <netdb.h>              /* hostent, gethostby*, getservby* */
#include <syslog.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <setjmp.h>

#include <sys/types.h>
//#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/time.h>

#define ISSERVER
#define MY_NAME "nbd-server"
#include "cliserv.h"

#ifdef USING_SSL
#include <openssl/lhash.h>
#include <openssl/bn.h>
#include "sslincl.h"
#endif


#ifdef USING_SSL
#define TEST_CERT       "server.pem"
#define BUFSIZE 16*1024
#endif

#ifdef USING_SSL
extern int MS_CALLBACK verify_callback(int ok, X509_STORE_CTX *ctx);
extern int set_cert_stuff(SSL_CTX *ctx, char *cert_file, char *key_file);
int init_ssl_connection(SSL *con);
#endif
  
/* lifted from fs.h */
#define BLKGETSIZE _IO(0x12,96) 


# define MAXCONN 8
# define MAXPATHLENGTH 1024
# define MAXFILE 32

# define NBD_LINEAR_MODE 'l'
# define NBD_STRIPE_MODE 0
# define NBD_MIRROR_MODE 1

#if defined(DEBUG) && DEBUG > 0
  #if DEBUG > 1
    #define DODBG
  #endif
  #undef DEBUG
  #define DEBUG( a... ) if(flags & F_DEBUG)fprintf(stderr, a ) /**/
  #define DEBUG_DFLT 1
#else
  #undef DEBUG
  #define DEBUG( a... )
  #define DEBUG_DFLT 0
#endif

#define F_READONLY 1
#define F_DEBUG    2

   int current = -1;               /* index pointer for forked copies */

   int nport = 0;                  /* number of ports */
   int nfile = 0;                  /* number of files */
   int mode  = NBD_LINEAR_MODE;    /* default linear raid */
   int blksize = 1024;             /* host device blocksize (default 1k) */
   int negotiate_timeout = 90l;    /* die and restart if not done in 90s */

   u64 exportsize;                 /* ...and its total length */
   static int pids[MAXCONN];       /* pid list for use by launch */
   int flags = 0;
   static char signature[NBD_SIGLEN];
                                   /* session id for exported file */
   static int has_sig = 0;         /* chosen a sig? */
                                   /* Files I'm exporting */
   u64 sizes[MAXFILE];             /* file sizes */
   int exports[MAXFILE];           /* file descriptors */
   static char exportnames[MAXFILE][MAXPATHLENGTH];
#ifdef USING_SSL
   static int using_ssl       = 0; /* amended later */
   static int verify_depth    = 0;
   static char *cipher        = NULL;
   static int s_server_verify = SSL_VERIFY_NONE;
   static char *s_cert_file   = TEST_CERT,
               *s_key_file    = NULL;
   static char *s_dcert_file  = NULL,
               *s_dkey_file   = NULL;
   static SSL_CTX *ctx        = NULL;
   static int s_server_session_id_context[MAXCONN];
   static int nocert          = 0;
   static char *CApath        = NULL,
               *CAfile        = NULL;
           BIO *bio_err       = NULL;
   SSL_METHOD  *meth          = NULL;
#endif

   struct nbd_server {
       int       i;                 /* PTB identifier 0 to MAXCONN-1 */
       int       port;              /* PTB port we listen to */
       int       pid;               /* PTB server process id */
       int       sock;              /* PTB descriptor of net connection */
       int       socket;            /* PTB principal socket for bind */
#ifdef USING_SSL
       SSL*      con;
#endif
   };
   static struct nbd_server servers[MAXCONN];
   static struct nbd_server manager = {
         -1,    // PTB id
         -1,    // PTB port
         -1,    // PTB pid
         -1,    // PTB sock
         -1,    // PTB socket
#ifdef USING_SSL
          0,    // PTB SSL *con
#endif
   };

   u64 lastpoint = -1;               /* PTB last seek position */
   static u64 lastpt[MAXFILE];       /* PTB per export component */
   int   lasti     = -1;             /* PTB last component sought */

   inline unsigned fileit(int direction, int fd, void *buf, unsigned len) {
      int res;
      unsigned count = 0;
      while (len > 0) {
         switch (direction) {
         case 0:
           DEBUG("*[%u@%Lu]", len, lastpoint);
           res = read(fd, buf, len);
           if (res <= 0) {
               perr("Read failed in readit: %m\n"); exit(1);
           }
           break;
         default:
         case 1:
           DEBUG("+[%u@%Lu]", len, lastpoint);
           res = write(fd, buf, len);
           if (res <= 0) {
               perr("Write failed in writeit: %m\n"); exit(1);
           }
           break;
         }
         len   -= res;
         buf   += res;
         count += res;
      }
      return count;
   }

#ifdef USING_SSL
   inline int ssl_check(SSL * con, int res) {

             switch (SSL_get_error(con,res)){ 
               case SSL_ERROR_WANT_WRITE:
               case SSL_ERROR_WANT_X509_LOOKUP:
               case SSL_ERROR_SYSCALL:
               case SSL_ERROR_ZERO_RETURN:
               case SSL_ERROR_SSL:
                 return -1;
              }
              return 0;
   }
#endif

   static fd_set fds, xfds;

   inline unsigned netit(struct nbd_server *self, int direction, void *buf, unsigned len)
   {
      int res;
      unsigned count = 0;

      while (len > 0) {

         struct timeval net_timeout = { negotiate_timeout, 0 };
                                // PTB. Urgle. must have 60 > sync_interval
         struct timeval * timeout = & net_timeout;
           

         switch (direction) {
         case 0:

           DEBUG("*");

#ifdef USING_SSL
           if (using_ssl) {
             res = SSL_read(self->con, buf, len);
             if (ssl_check(self->con, res) < 0)
               exit(1);
           } else
#endif
           {
             fd_set * rfds = & fds, * efds = & xfds;

             FD_ZERO(efds); FD_SET(self->sock,efds);
             FD_ZERO(rfds); FD_SET(self->sock,rfds);

             if (select(self->sock+1,rfds,NULL,efds,timeout) <= 0) {
               perr("Read timed out in readit: %m\n"); exit(1);
             }
             res = read(self->sock, buf, len);
           }
           if (res < 0) {
               perr("Read failed in readit: %m\n"); exit(1);
           }
           break;
         default:
         case 1:
           DEBUG("+");
#ifdef USING_SSL
           if (using_ssl) {
             res = SSL_write(self->con, buf, len);
             if (ssl_check(self->con, res) < 0)
               exit(1);
           } else
#endif
           {
             fd_set * wfds = & fds, * efds = & xfds;

             FD_ZERO(efds); FD_SET(self->sock,efds);
             FD_ZERO(wfds); FD_SET(self->sock,wfds);
             
             if (select(self->sock+1,NULL,wfds,efds,timeout) <= 0) {
               perr("Write timed out in writeit: %m\n"); exit(1);
             }
             res = write(self->sock, buf, len);
           }
           if (res <= 0) {
               perr("Write failed in writeit: %m\n"); exit(1);
           }
           break;
         }
         len   -= res;
         buf   += res;
         count += res;
      }
      return count;
   }

   inline int readit(struct nbd_server *self, void *buf, int len)
   {
      return netit(self,0,buf,len);
   }

   inline int writeit(struct nbd_server *self, void *buf, int len)
   {
      return netit(self,1,buf,len);
   }



   void usage() {
         printf("Usage: port [port ..] file_to_export .. [size] "
                "[-l|-0] [-r] [-i sig] [-b blksiz] [-t timeout]"
#ifdef USING_SSL
                "[-cert certfile] [-key keyfile] "
                "[-CApath CApath] [-CAfile CAfile] [-verify depth]\n"
#endif
               "        -r read only\n"
               "        if port is set to 0, stdin is used (for running from inetd)\n"
               "                address of machine trying to connect\n" );
    }

   int cmdline(int argc, char *argv[])
   {
      int i = 0;
   
      if (argc < 3) 
        return 1;
   
      for (i = 1; i < argc ; i++) {
         int port = 0;
         char *strerr;
         port = strtol(argv[i],&strerr, 10);
         if (!*argv[i] || *strerr)
            break;
         servers[nport++].port = port;
      } 
   
      for ( ; i < argc; i++) {
        if (*argv[i] == '-') 
          break;
        strncpy(exportnames[nfile++],argv[i],MAXPATHLENGTH);
      }

      if (DEBUG_DFLT > 0)
         flags |= F_DEBUG;
   
      for ( ; i < argc; i++) {
         if (*argv[i] == '-') {
            switch (argv[i][1]) {
            case 'l':
               mode = NBD_LINEAR_MODE;
               break;
            case '0':
               mode = NBD_STRIPE_MODE;
               break;
            case '1':
               mode = NBD_MIRROR_MODE;
               break;
            case 'r':
               flags |= F_READONLY;
               break;
            case 'd':
               flags |= F_DEBUG;
               break;
            case 'i':
               if (++i >= argc) 
                 return 1;
               strncpy(signature, argv[i], sizeof(signature));
               has_sig = 1;
               break;
            case 'b':
               if (++i >= argc) {
                 return 1;
               } else {
                 char *strerr;
                 blksize = strtol(argv[i], &strerr, 10);
                 if (!*argv[i] || *strerr) {
                   return 1;
                 }
               }
               break;
              case 't':
               if (i+1 >= argc) 
                    return 1;
               else {
                    char *strerr;
	            i++;
                    negotiate_timeout = strtol (argv[i], &strerr, 10);
                    if (!*argv[i] || *strerr) {
                      negotiate_timeout = 60l; /* PTB - restore default */
	              return 1;
                    }
                    MSG("client says negotiation timeout is %d\n",
                      negotiate_timeout);
	            continue;
               } 
               break;
#ifdef USING_SSL
              case 'c':
                if (i+1 >= argc)
                  return 1;
		if (!strcmp(argv[i], "-cert")){
		  s_cert_file = argv[++i];
                  using_ssl = 1;
		  MSG("server cert file is %s\n",s_cert_file);
                break;
		}
		if (!strcmp(argv[i], "-cipher")){
		  cipher = argv[++i];
                  using_ssl = 1;
		  MSG("client cipher: %s\n",cipher);
		  break;
		}
		return 1;
              case 'k':
                if (i+1 >= argc)
                  return 1;
		if (strcmp(argv[i], "-keyfile"))
                    return 1;
		s_key_file = argv[++i];
                using_ssl = 1;
		MSG("server key cert file is %s\n",s_key_file);
                break;
              case 'C':
                  if (i+1 >= argc)
                    return 1;
		  if (!strcmp(argv[i], "-CApath")){
                    CApath = argv[++i];
                    using_ssl = 1;
                    MSG("client CApath is %s\n",CApath);
                    break;
                  }
		  if (!strcmp(argv[i], "-CAfile")){
                    CAfile = argv[++i];
                    using_ssl = 1;
                    MSG("client CAfile is %s\n",CAfile);
                    break;
                  }
                  return 1;
              case 'v':
                if (i+1 >= argc)
                  return 1;
		if (!strcmp(argv[i], "-verify")){
                  char *err;
                  i++;
                  verify_depth = strtol (argv[i], &err, 10);
		  s_server_verify = SSL_VERIFY_PEER;
                  if (!*argv[i] ||  *err)
                      return 1;
                    using_ssl = 1;
                    MSG("client says verify_depth is %d\n",verify_depth);
                    break;
                } 
                return 1;
#endif
            default:
              return 1;
            }
         } else {
            int es = atoi(argv[i]);
            exportsize = es;
         }
      }
      return 0;
      
   }

   int connectme(struct nbd_server * self)
   {
      struct sockaddr_in addrin;
      int addrinlen = sizeof(addrin);
      int flag = 1;
      int size = 1;
      int err  = 0;
      struct timeval timeout = { tv_sec: 60, tv_usec: 0 };
      struct linger lin = { l_onoff: 0, l_linger: 0 };

      if (self->socket > 0) {
        shutdown(self->socket,2);
        close(self->socket);
        self->socket = -1;
      }
      if (self->socket <=  0)
        self->socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
      if (self->socket < 0) {
        MSG("socket: %m\n");
        return -1;
      }
   
      DEBUG("Server waiting for connections... ");
      addrin.sin_family = AF_INET;
      addrin.sin_port = htons(self->port);
      addrin.sin_addr.s_addr = 0;

      if (setsockopt(self->socket, SOL_SOCKET, SO_REUSEADDR, (void *)&flag, sizeof(flag)) < 0) {
        MSG("warning: setsockopt REUSEADDR failed with %m\n");
      }
      if (setsockopt(self->socket, SOL_SOCKET, SO_KEEPALIVE, &size, sizeof(size)) < 0) {
        MSG("warning: setsockopt KEEPALIVE failed with %m\n");
      }
      if (setsockopt(self->socket, SOL_SOCKET, SO_RCVTIMEO, &timeout, sizeof(timeout)) < 0) {
        MSG("warning: setsockopt RCVTIMEO failed with %m\n");
      }
      if (setsockopt(self->socket, SOL_SOCKET, SO_LINGER, &lin, sizeof(lin)) < 0) {
        MSG("warning: setsockopt LINGER failed with %m\n");
      }

      DEBUG("bind, ");
      if (bind(self->socket, (struct sockaddr *) &addrin, addrinlen) < 0) {
        MSG("bind: %m\n");
        err = -2;
        goto fail;
      }
      DEBUG("listen, ");
      if (listen(self->socket, 1) < 0) {
        MSG("listen: %m\n");
        err = -3;
        goto fail;
      }

      DEBUG("accept, ");
      self->sock = accept(self->socket, (struct sockaddr *) &addrin, &addrinlen);
      if (self->sock < 0) {
        MSG("accept: %m\n");
        err = -4;
        goto fail;
      }

      if (fcntl(self->sock,F_SETFL,O_NONBLOCK) < 0) {
        MSG("warning: fcntl NONBLOCK failed with %m\n");
      }

      printf("server opened port #%d\n",self->port);
      return self->sock;

fail:
      shutdown(self->socket,2);
      close(self->socket);
      self->socket = -1;
      return err;

   }

   void closing(struct nbd_server *self){
     MSG ("server closing port #%d\n",self->port);
     DEBUG ("server closing port #%d\n",self->port);
     if (self->sock > 0) {
       shutdown(self->sock,2);
       close(self->sock);
     }
     self->sock = -1;
#ifdef USING_SSL
     if (using_ssl)  {
       SSL_set_shutdown(self->con,SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
#ifndef TRY_SSL_RECONNECT
       if (self->con)
         SSL_free(self->con);
       self->con = 0;
#endif
     }
#endif

   }
     
   void closenet(int x){
      // PTB this is a closing() without the benefit of pointer to self.
      struct nbd_server * self = & servers[current];
      closing(self);
   }

   static inline s64 nbd_seek(int fildes, s64 offset, int whence)
   {
#ifdef HAVE_LSEEK64
       return lseek64(fildes, (loff_t) offset, whence);
#elif HAVE_LLSEEK
       return llseek(fildes, (loff_t)offset, whence);
#else
       if ((offset >> 31) != 0) {
          PERR("Cannot seek to more than 2^31-1 (%Lx) within one file!\n", a);
          return (s64) -1;
       }
       return (s64) lseek(fildes, (off_t) offset, whence);
#endif
   }


#define SEND writeit(&servers[current], &reply, sizeof( reply ));
#define ERROR { reply.error = htonl(-1); SEND; reply.error = 0; lastpoint = -1; }

   int maybeseek(u64 a)
   {
      int i = lasti; 

      if (a >= exportsize) {
        perr("Cannot happen\n"); exit(1);
      }

      DEBUG("@%Lu",a);

      if (lastpoint == a) {    // PTB fast path
        DEBUG(":%d/%Lu/%Lu",lasti,lastpt[lasti]/blksize,lastpt[lasti]%blksize);
        return exports[lasti]; 
      }
      lastpoint = a;          // PTB setup fast path for next time, offset below

      switch (mode) {
           u64 b;

           // PTB choose the fd to write to

           case NBD_LINEAR_MODE:
           for (i = 0; i < nfile; i++) {
             u64 size = sizes[i];
             if (a < size)
               break;
             a -= size;
           }
           if (i >= nfile) {
             perr("Cannot happen to me!\n"); exit(1);
           }
           break;
           case NBD_MIRROR_MODE:
             i = 0;
           break;
           case NBD_STRIPE_MODE:
             b = a % blksize;   // offset within block
             a/= blksize;       // block number
             i = a % nfile;     // file id holding that block number
             a/= nfile;         // block number within file
             a*= blksize;
             a+= b;             // plus offset within block
           break;
      }

      if (lastpt[i] > 0 && lastpt[i] == a) {
                           // PTB medium fast path
          lasti = i;
          DEBUG(":%d/%Lu/%Lu",i,a/blksize,a%blksize);
          return exports[lasti];
      }
                            // PTB slow path

      switch (mode) {
        int j;
        case NBD_STRIPE_MODE:
        case NBD_LINEAR_MODE:
          if (nbd_seek(exports[i], a, SEEK_SET) < 0) {
             perr("Can not seek locally!\n");
             exit(1);
          }
        break;
        case NBD_MIRROR_MODE:
          for (j = 0; j < nfile; j++) {
            if (nbd_seek(exports[j], a, SEEK_SET) < 0) {
               perr("Can not seek locally!\n");
               exit(1);
            }
          }
        break;
      }
      lasti = i;
      lastpt[lasti] = a;

      DEBUG(":%d/%Lu/%Lu",i,a/blksize,a%blksize);
      return exports[lasti];
   }

   unsigned opfile(int direction, char * buf, u64 foffset, unsigned len) {

            // split into sections of at most blocksize

            int boffset = 0;  // current offset within buf
            int tot     = 0;  // total bytes transfered so far

            while (len > 0) {

              int blockoffset = foffset % blksize; // offset within block
              unsigned sublen = 0;                 // transfer this much
              int export = -1;
              int count  = -1;
             
              // PTB cut sublen down to go up to block boundary only

              if ( len + blockoffset > blksize )
                sublen = blksize - blockoffset;
              else
                sublen = len;

              // PTB this sets lasti, lastpoint, lastpt[lasti]
              export   = maybeseek(foffset);

              switch(mode) {
              default:
              // NBD_LINEAR_MODE
              // NBD_STRIPE_MODE
                count = fileit(direction, export, buf + boffset, sublen);
                break;
              case NBD_MIRROR_MODE:
                switch(direction) {
                int j; 
                fd_set fds, * rfds;
                int fmax;

                case 0: // read

                  rfds = & fds; 
                  fmax = 0;
                  FD_ZERO(rfds);
                  for (j = 0; j < nfile; j++) {
                    int fd = exports[j];
                    fmax = fmax < fd ? fd : fmax;
                    FD_SET(fd, rfds);
                  }

                  if (select(fmax+1,rfds,NULL,NULL,NULL) <= 0)
                    return tot;

                  for (j = 0; j < nfile; j++) {
                    int fd = exports[j];
                    if (FD_ISSET(fd, rfds)){
                      export = fd;
                      break;
                    }
                  }

                  count = fileit(direction, export, buf + boffset, sublen);
                  break;
                case 1: // write
                  for (j = 0; j < nfile; j++) {
                    count = fileit(direction, exports[j], buf+boffset, sublen);
                    if (count <= 0)
                      return tot;
                  }
                  break;
                }
                break;
              }

              if (count <= 0)
                return tot;
              lastpoint += count;
              lastpt[lasti] += count;

              // PTB check if we can use the same seek point next time

              switch(mode) {
              case NBD_LINEAR_MODE:
                if (lastpt[lasti] >= sizes[lasti]) {
                  lastpoint = -1; // PTB invalidate maybeseek shortcut
                }
                break;
              case NBD_MIRROR_MODE:
                break;
              case NBD_STRIPE_MODE:
                if (count + blockoffset >= blksize) {
                  lastpoint = -1; // PTB invalidate maybeseek shortcut
                }
                break;
              }

              foffset += count;
              boffset += count;
              len     -= count;
              tot     += count;
            }
            return tot;
   }

   unsigned writefile(char * buf,u64 foffset, unsigned len) {
            return opfile(1,buf,foffset,len);
   }
   unsigned readfile(char * buf,u64 foffset, unsigned len) {
            return opfile(0,buf,foffset,len);
   }

   int mainloop(struct nbd_server *self)
   {
      struct nbd_request request;
      struct nbd_reply reply;
   
      DEBUG("Entering request loop!\n");
      reply.magic = htonl(NBD_REPLY_MAGIC);
      reply.error = 0;

      while (1) {

         char buf[NBD_MAX_SECTORS*512 + 1000 + NBD_MAXCONN] ;
         int len = 0;
      
      #ifdef DODBG
         static int i;
         i++;
         printf("%d.%d: ", self->port, i);
      #endif
      
      readit(self, &request, sizeof(request));

      #if 1 /* PTB */
         switch (sizeof(request.from)) {
         case 4: request.from = ntohl(request.from); 
            break;
         case 8: request.from = ntohll(request.from); 
            break;
         default: 
            perr("Impossible from field size: %m\n");
            closing(self);
            return -1;
         }
      #else
         request.from = ntohll((u64) request.from);
      #endif

         request.type = ntohl(request.type);
         len = ntohl(request.len);
      
         if (request.magic != htonl(NBD_REQUEST_MAGIC)) {
           perr("Not enough magic in packet.\n"); 
           continue;       /*  PTB now drop instead of erroring out */
         }
      #ifdef DODBG
         printf("%s from %d (%d) len %d, ", (request.type ? "WRITE" : "READ"),
               (int) request.from, (int) request.from / 512, len);
      #endif
         memcpy(reply.handle, request.handle, NBD_HANDLE_LENGTH);
         if ( request.from + len > exportsize) {
            DEBUG("[RANGE! (%Lu+%Lu)]",
                     exportsize,(request.from + len) - exportsize);
            ERROR;
            continue;
         }
         if ((flags & F_READONLY) && request.type == 1) {
            DEBUG("[TYPE! (0x%x)]", (unsigned)request.type);
            ERROR;
            continue;
         }
         switch (request.type & 0x03) {

         case 1:  /* WRITE */
            DEBUG("wr: net->buf, ");
            readit(self, buf, len);
            if (writefile(buf,request.from,len) != len) {
               DEBUG("Write failed in mainloop: %m");
               ERROR;
               continue;
            }
            DEBUG("buf->exp, ");
            SEND;
            DEBUG("OK!\n");
            break;

         case 2: /* KEEPALIVE */
         case 3: /* KEEPALIVE */
            DEBUG("exp->buf, ");
            // embed a payload sig equal to offset mod 256 repeated len times
            memset (buf + sizeof(reply), request.from % 256 , len);

            DEBUG("buf->net, ");
            reply.magic = ntohl(NBD_REPLY_MAGIC);
            memcpy(buf, &reply, sizeof(reply));
            writeit(self, buf, sizeof(reply) + len);
            DEBUG("OK!\n");
            break;
         default:
          case 0: /* READ */
            DEBUG("exp->buf, ");
            if (readfile(buf + sizeof(reply), request.from, len) != len) {
               DEBUG("Read failed in mainloop: %m");
               ERROR;
               continue;
            }

            DEBUG("buf->net, ");
            memcpy(buf, &reply, sizeof(reply));
            writeit(self, buf, sizeof(reply) + len);
            DEBUG("OK!\n");
            break;
         }
         continue;

      }
   }

  int getsig(char * buf, int n) {
        int u;
        u = open("/dev/random",O_RDONLY);
        if (u < 0)
          return u;
        if (read(u, buf, n) < n) {
          close(u);
          return -1;
        }
        return 0;
  }

#ifdef USING_SSL
  int init_ssl_connection(SSL *con)
  {
        int i;
        const char *str;
        long verify_error;
        MS_STATIC char buf[BUFSIZE];
  	DEBUG("Negotiation of SSL connection...\n");

        if ((i=SSL_accept(con)) <= 0) {
              MSG("ERROR en init_ssl_connection\n");
              DEBUG("ERROR en init_ssl_connection\n");
              verify_error = SSL_get_verify_result(con);
              if (verify_error != X509_V_OK) {
                      MSG("verify error:%s\n",
                              X509_verify_cert_error_string(verify_error));
              } else
                      ERR_print_errors(bio_err);
  	      DEBUG("SSL negotiation failed!\n");
              return 0;
        }


        if (SSL_get_shared_ciphers(con,buf,BUFSIZE))
                 DEBUG("Shared ciphers: %s\n",buf);
        str = SSL_CIPHER_get_name(SSL_get_current_cipher(con));
        DEBUG("CIPHER is %s\n",(str)?str:"(NONE)");
        if (con->hit){
          DEBUG("Reconnection...");
#ifdef TRY_SSL_RECONNECT
          // PTB - we should try this anyway. Looks harmless.
#endif
          SSL_renegotiate(con);
          i = SSL_do_handshake(con);
          DEBUG("%s... Reused session-id\n",
              (i>=0)?"Good handshake!":"Bad handshake"); 
          if (i < 0) {                          // PTB add
  	      DEBUG("SSL negotiation failed!\n");
              return 0;
          }
        }
        if (SSL_ctrl(con,SSL_CTRL_GET_FLAGS,0,NULL) &
                TLS1_FLAGS_TLS_PADDING_BUG) {
                MSG("Peer has incorrect TLSv1 block padding\n");
        }

 	DEBUG("SSL connection ready!");
        return 1;
  }
  
  int init_server_ssl_stuff() {
    if (nocert)
      {
	s_cert_file  = NULL;
	s_key_file   = NULL;
	s_dcert_file = NULL;
	s_dkey_file  = NULL;
      }

    DEBUG("Initializing SSL context\n");

    meth = SSLv23_server_method();

    SSL_load_error_strings();
    SSLeay_add_ssl_algorithms();
    
    if (! bio_err)
      bio_err = BIO_new_fp(stderr,BIO_NOCLOSE);
    
    ctx = SSL_CTX_new(meth);
    if (! ctx)
      {
        ERR_print_errors(bio_err);
        goto end;
      }
    
    SSL_CTX_set_quiet_shutdown(ctx,1);
    SSL_CTX_set_options(ctx,0);
    SSL_CTX_sess_set_cache_size(ctx,128);
    
    if (!SSL_CTX_load_verify_locations(ctx,CAfile,CApath) ||
        !SSL_CTX_set_default_verify_paths(ctx))
      {
        ERR_print_errors(bio_err);
        /* goto end; */
      }
        
    DEBUG("  loading cert from file %s\n", s_cert_file);

    if (!set_cert_stuff(ctx,s_cert_file,s_key_file))
      goto end;
    if (s_dcert_file)
      {
        if (!set_cert_stuff(ctx,s_dcert_file,s_dkey_file))
          goto end;
      }

    DEBUG("  preferred cipher is (%s)\n",cipher);

    if (cipher)
      SSL_CTX_set_cipher_list(ctx,cipher);
    SSL_CTX_set_verify(ctx,s_server_verify,verify_callback);
    SSL_CTX_set_session_id_context(ctx,(void*)&s_server_session_id_context,
                                   sizeof s_server_session_id_context);

    SSL_CTX_set_client_CA_list(ctx,SSL_load_client_CA_file(CAfile));

    DEBUG("... SSL initialized\n");
    
    return 0;

  end:
    DEBUG("Error in SSL initialization\n");
    if (ctx) SSL_CTX_free(ctx);
    return -1;

  }

#endif 

  void resetsighandlers() {
       // PTB all signals to defaults
       int k = 0;
       for (k = 1; k < 30; k++) {
#if defined(__GLIBC__) && __GLIBC__ >= 2
         struct sigaction sa = { {SIG_DFL}, {{0}}, SA_ONESHOT, NULL };
#else
         struct sigaction sa = { {SIG_DFL}, 0, SA_ONESHOT, NULL };
#endif
         sigfillset(& sa.sa_mask);
         sigaction(k, & sa, NULL);
       }
  }                    

  static jmp_buf * trampoline = 0;

  inline void bounce(int x) {
     jmp_buf * j = trampoline;
     trampoline = 0;
     if (j)
       longjmp(*j, x);
  }

  inline void set_alarm(unsigned long timeout) {
         alarm(timeout);
         if (timeout > 0) {
           signal(SIGALRM, bounce);
           // PTB a SIGALRM will make us bounce on the trampoline
         } else {
           // PTB ignore SIGALRM
           signal(SIGALRM, SIG_IGN);
         }
   }

  int negotiate(struct nbd_server *self) {

         void (*sig)(int);
         u64 size_host = 0;
         unsigned long blksize_host = 1024;
         int err = 0;
         jmp_buf my_jmp;
         unsigned long timeout = negotiate_timeout;

 	 DEBUG("Server begins negotiation\n");
         resetsighandlers();

         trampoline = &my_jmp;
         if (setjmp(my_jmp)) {  
 	   DEBUG("Server timed out in negotiate\n");
           err = -15;
           goto fail;
         }

         if (self->sock > 0) {
 	   DEBUG("Server closes pending connections\n");
           closing(self);
         }
    
 	 DEBUG("Server sets timer for negotiation\n");
         set_alarm(timeout);

         // if port==0 we use stdin as socket, i.e. sock=0
         if (self->port > 0)
           connectme(self);
         else
           self->sock = 0;

         if (self->sock < 0) {
 	   DEBUG("Server failed connectme\n");
           err = -1;
           goto fail;
         }

         if (self->sock > 0)
           setmysockopt(self->sock);

#ifdef USING_SSL
         if (using_ssl) {

 	   DEBUG("Socket connected, now let's initialize the SSL connection\n");

             if (!self->con)  // make new ssl
               self->con = (SSL *) SSL_new(ctx);

             if (!SSL_clear(self->con))
               DEBUG("SSL_clear returned 0, are we using session id?\n");
  
             SSL_set_accept_state(self->con);
  
             if  (!SSL_set_fd(self->con,self->sock)
               || !init_ssl_connection(self->con)) {
                DEBUG("SSL session uninitialized\n");
                err = -1;
                goto fail;
             }
         
             DEBUG("SSL connection initialized\n");
         }
#endif

 	 DEBUG("Setting signal handlers\n");

         sig = signal(SIGTERM,closenet);
         sig = signal(SIGINT, closenet);

  	 DEBUG("Go on with the nbd negotiation: ");
  
  	 DEBUG("passwd... ");
  
         if (writeit(self, INIT_PASSWD, 8) < 0) {
            perr("Negotiation failed: %m\n");
            err = -3;
            goto fail;
         }

  	 DEBUG("OK!, cliserv_magic... ");

         cliserv_magic = htonll(cliserv_magic);
         if (writeit(self, &cliserv_magic, sizeof(cliserv_magic)) < 0) {
            perr("Negotiation failed: %m\n");
            err = -4;
            goto fail;
         }
 
 	 DEBUG("OK!, size... ");
 
         size_host = htonll(exportsize);
         if (writeit(self, &size_host, sizeof(size_host)) < 0) {
            perr("Negotiation failed: %m\n");
            err = -5;
            goto fail;
         }
 
 	 DEBUG("OK!, signature... ");
 
         if (writeit(self, signature, NBD_SIGLEN) < 0) {
            perr("Negotiation failed: %m\n");
            err = -6;
            goto fail;
         }
 
 	 DEBUG("OK!, blksize... ");
 
         if (readit(self, &blksize_host, sizeof(blksize_host)) < 0) {
            perr("Negotiation failed: %m\n");
            err = -7;
            goto fail;
         }
         if (blksize_host > blksize)
            blksize = blksize_host;
         if (writeit(self, &blksize, sizeof(blksize)) < 0) {
            perr("Negotiation failed: %m\n");
            err = -8;
            goto fail;
         }                   

 	 DEBUG("OK!\n Negotiation done, disabling timer and return\n");
         set_alarm(0l);
         return err = 0;
fail:
 	 DEBUG("Negotiation failed, sock descriptor is %d\n", self->sock);
         closing(self);
 	 DEBUG("Negotiation failed, disabling timer and return\n");
         set_alarm(0l);
         return err;
  }

  int launch(struct nbd_server * self) {

         int pid = -1;
         int err = 0;
         int i = self->i;
         struct timeval timeout = {5,0};
         int j = 0;
   
         // detach
         pid = fork();

         switch(pid) {
         case 0:  // PTB child continues
            break;
         case -1: // PTB fork error
            return self->pid = -1;
         default: // PTB parent returns with pid
            return self->pid = pid;
         }

         // child
         current = i;
         pids[self->i] = self->pid = getpid();
         self->sock = -1; // PTB invalidate current descriptor

         // PTB copy instead of share the resource FDs
         for (j = 0; j< nfile; j++) {
             int export = exports[j];
             char * exportname = exportnames[j];
             FILE * f = fdopen(export,(flags&F_READONLY) ?  "r" : "r+"); 
             if (!f)
 	        DEBUG("Cannot reopen file descriptor %d\n", export);
             if (!freopen(exportname, (flags&F_READONLY) ?  "r" : "r+", f))
 	        DEBUG("Cannot reopen file %s\n", exportname);
         }

         while(1) {
           // PTB repeatedly try to set up initial connection
           struct timeval timeout = {10,0};

           while(1) { 
             struct timeval timeout = {10,0};

             err = negotiate(self);
             if (err >= 0) 
             // PTB success
                 break;

 	     DEBUG("Server tries to sleep 10\n");
             select(0,NULL,NULL,NULL,&timeout);
           }

           // do protocol
           err = mainloop(self);

           if (err >= 0) 
           // PTB success = planned termination
               break;
           
 	   DEBUG("Server tries to sleep 10\n");
           select(0,NULL,NULL,NULL,&timeout);
           exit(2);
         }

         /* child death - ptb */

         // PTB attempt delay to avoid fast SIGCHLDs in sequence
 	 DEBUG("Server tries to sleep 5\n");
         select(0,NULL,NULL,NULL,&timeout);
         return err;
  }

       
  u64 getsize1(int fd) {
         u64 size = 0;

         if (!size) {
               off_t es = 0;
               DEBUG("looking for export size with lseek SEEK_END\n");
               if ((es = lseek(fd, 0, SEEK_END)) != -1 && es != 0) {
                  size = es;
               }
         }
         if (!size) {
               off_t es = 0;
               struct stat stat_buf = { 0,0,0,0,0,0,0,0,0,0,0,0,0, } ;
               int error;
               DEBUG("looking for export size with fstat\n");
               if ((error = fstat(fd, &stat_buf)) != -1
                && (es = stat_buf.st_size) != 0 ) { 
                  size = es;
               }
         }
         if (!size) {
               off_t es = 0;
               DEBUG("looking for export size with ioctl BLKGETSIZE\n");
               if( ! ioctl(fd, BLKGETSIZE, &es) && es != 0) {
                  size = es;
                  size <<= 9; /* assume blocksize 512 */
               }
         }
         size = ( size / blksize ) * blksize;
         return size;
  }

  u64 getsize(int n, int * fds, u64 * sizes) {
     int i = 0;
     u64 totalsize = 0;

     for (i = 0; i < n; i++) {
         u64 size = 0;
         int fd = fds[i];

         size = getsize1(fd);

#if !defined(HAVE_LSEEK64) && !defined(HAVE_LLSEEK)
        if ((size >> 31) != 0) {
            PERR("%s is too large to export!\n", exportnames[i]);
            exit(1);
        }
#endif

         sizes[i] = size;
         switch (mode) {
           case NBD_LINEAR_MODE: totalsize += size;
                   break;
           case NBD_MIRROR_MODE: totalsize =
             (i<=0||totalsize<=0||size<totalsize) ? size : totalsize;
                   break;
           case NBD_STRIPE_MODE: totalsize =
             (i<=0||totalsize<=0||size<totalsize/i)? size*(i+1) :
                                                        (totalsize/i)*(i+1);
                   break;
         }
     }
     return totalsize;
  } 

  int restart(int died) {
 
    // check and restart a known pid. (or all if pid = -1)
 
    int i = 0;
    int count = 0;
 
    for ( i = 0; i < nport; i++)  {
 
         int pid  = pids[i];
 
         MSG("main childminder checking pid %d\n", pid);
 
         if ( pid < 0 || kill(pid,0) < 0 || died == pid )  {
 
             pids[i] = -1;
 
             pid = launch(&servers[i]);
             if (pid > 0) {
               pids[i] = pid;
               count ++;
             }
         }
    }
    return count;
  }

  void sigchldhandler() {
 
    // reap and restart dead children.
 
    int died = 0;
    int count = 0;
 
    while ( died = wait4(-1,0,WNOHANG,0), died > 0 ) 
      count += restart(died) ;
 
    if (count <= 0)
      // an unknown child died so check all
      count += restart(-1) ;
  }                                

  void propagate(int n) {
            int i;
            for (i = 0; i < nport; i++) {
              int pid = pids[i];
              if (pid > 0)
                kill(pid,n); // renegotiate
            }
  }

  void sighandler(int n) {
 
        switch (n) {
        case SIGHUP:  // PTB ignore
            propagate(n);
            MSG( "sighandler propagates exit %d in manager\n", n);
            break;
        case SIGTTIN:
        case SIGTTOU:
        case SIGPIPE:
            MSG( "sighandler ignores exit %d in manager\n", n);
            break;
        case SIGCHLD:
            MSG( "sighandler relaunches child from manager \n");
            sigchldhandler();
            break;
        case SIGTERM:
        case SIGINT:
            propagate(n);
        case SIGSEGV:
            exit(0);
            MSG( "sighandler terminates manager safely\n");
        default:     // PTB ignore 
            MSG( "sighandler ignores exit %d in manager\n", n);
            break;
        }                
  }            


  void setsighandlers() {
       // PTB handle all signals in one sighandler
       int k = 0;
       for (k = 1; k < 30; k++) {
#if defined(__GLIBC__) && __GLIBC__ >= 2
         struct sigaction sa = { {sighandler}, {{0}}, SA_RESTART, NULL };
#else
         struct sigaction sa = { {sighandler}, 0, SA_RESTART, NULL };
#endif
         sigfillset(& sa.sa_mask);
         sigaction(k, & sa, NULL);
       }
  }                    

  void init(struct nbd_server * self, int i) {
      self->i = i;
  }

  int main(int argc, char *argv[]) {

      int i, npids=0;
      struct nbd_server * self = & manager;
   
      logging();
      if (cmdline(argc, argv)) {
        usage();
        exit(0);
      }

      if (!has_sig) 
        has_sig = !getsig(signature, NBD_SIGLEN);

      for (i = 0; i < nfile; i++) {
         char * exportname = exportnames[i];
         int export = -1;
         export = open(exportname, (flags&F_READONLY) ?  O_RDONLY : O_RDWR);
         if (export < 0) {
            perr("Could not open exported file %s: %m\n", exportname);
            exit(1);
         }
         exports[i] = export;
      }
         
      exportsize = getsize(nfile,exports,sizes);
      if (exportsize <= 0) {
         perr("Could not find size of exported block device: %m\n");
         exit(1);
      }

      syslog(LOG_INFO, "size of exported file/device is %Ld", exportsize);

      // detach
      if (fork())
        return 0;

#ifdef USING_SSL
      if (using_ssl && init_server_ssl_stuff() < 0)
        return -1;
#endif
   
      for (i = 0; i< nport; i++) {

         int pid = -1;

         init(&servers[i], i);

         pid = launch(&servers[i]);

         if (pid > 0)  {
           pids[i] = pid;
           npids ++;
         }
      }

     self->sock = -1;
      
     setsighandlers();          // PTB handle CHLD USR1 HUP TERM
     MSG("manager set signal handlers\n");                  

      while (npids > 0)
        sleep(5);
   
      return 0;
  }



